#include <Windows.h>
#include <commctrl.h>
#include <tchar.h>
#include <math.h>
#include "resource.h"
#pragma comment(lib, "comctl32.lib")
#define W4K 3840
#define H4K 2160
#define W8K (W4K * 2)
#define H8K (H4K * 2)
int width = W4K;
int height = H4K;
int scale = 2;
int colorconv = 0;
BITMAPINFO	bmpInfo;
LPBYTE		lpPixel = NULL, lpScanline = NULL;
TCHAR		szTempBuf[MAX_PATH];

typedef struct{
	int num;
	int frm;
	int len;
	int step;
	int isdpx;
	TCHAR szFileName[MAX_PATH];
	LPTSTR szBaseName;
	LPTSTR list;
} FILELIST;
FILELIST filelist;

void InitDIB(void) {

	if (lpPixel != NULL) GlobalFree(lpPixel);
			
	ZeroMemory(&bmpInfo, sizeof(bmpInfo));
	bmpInfo.bmiHeader.biSize	= sizeof(BITMAPINFOHEADER);
	bmpInfo.bmiHeader.biWidth	= (width >> scale);
	bmpInfo.bmiHeader.biHeight	= (height >> scale);
	bmpInfo.bmiHeader.biPlanes = 1;
	bmpInfo.bmiHeader.biBitCount = 32;
	bmpInfo.bmiHeader.biCompression = BI_RGB;

	lpPixel = (LPBYTE)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
		bmpInfo.bmiHeader.biWidth * bmpInfo.bmiHeader.biHeight * sizeof(RGBQUAD));

	if (lpScanline == NULL) {
		lpScanline = (LPBYTE)HeapAlloc(GetProcessHeap(),
			HEAP_ZERO_MEMORY, W8K * 6);
	}

	return;
}

BOOL QueryFileName(HWND hwnd)
{
	static OPENFILENAME ofn;
	static TCHAR szFilter[] = TEXT("DPX/RGB Files (*.dpx, *.[rgb])\0*.dpx;*.r;*.g;*.b\0All Files (*.*)\0*.*\0");
	ofn.lStructSize			= sizeof(OPENFILENAME);
	ofn.hwndOwner			= hwnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= szFilter;
	ofn.lpstrCustomFilter	= NULL;
	ofn.nMaxCustFilter		= 0;
	ofn.nFilterIndex		= 0;
	ofn.lpstrFile			= filelist.szFileName;
	ofn.nMaxFile			= MAX_PATH;
	ofn.lpstrFileTitle		= NULL;
	ofn.nMaxFileTitle		= 0;
	ofn.lpstrInitialDir		= NULL;
	ofn.lpstrTitle			= NULL;
	ofn.Flags				= OFN_FILEMUSTEXIST | OFN_READONLY;
	ofn.nFileOffset			= 0;
	ofn.nFileExtension		= 0;
	ofn.lpstrDefExt			= NULL;
	ofn.lCustData			= 0L;
	ofn.lpfnHook			= NULL;
	ofn.lpTemplateName		= NULL;
	return (GetOpenFileName(&ofn));
}

int GetFileList()
{
	HANDLE	hFind;
	WIN32_FIND_DATA	FindData;
	LPTSTR lpList;
	int k;

	filelist.isdpx = 1;
	filelist.num = filelist.len = 0;
	_tcscpy_s(szTempBuf, MAX_PATH, filelist.szFileName);
    filelist.szBaseName = filelist.szFileName;
	for (k = _tcslen(szTempBuf) - 1; k >= 0; k--) {
		if (szTempBuf[k] == TEXT('\"')) szTempBuf[k] = TEXT('\0');
		if (szTempBuf[k] == TEXT('\\')) {
			filelist.szBaseName = filelist.szFileName + k + 1;
			break;
		}
		if (szTempBuf[k] == TEXT('.')) {
			if (szTempBuf[k + 2] == TEXT('\0') &&_tcschr(TEXT("rgbRGB"), szTempBuf[k + 1])) {
				filelist.isdpx = 0;
			}
			while (k > 0 && _istdigit(szTempBuf[--k])) {
				szTempBuf[k] = TEXT('?');
			}
		}
	}

	hFind = FindFirstFile(szTempBuf, &FindData);
	if (hFind == INVALID_HANDLE_VALUE) return (-1);
	do {
		filelist.num++;
		k = _tcslen(FindData.cFileName);
		if (k > filelist.len) {
			filelist.len = k;
		}
	} while (FindNextFile(hFind, &FindData));
	FindClose(hFind);

	if (filelist.list != NULL) GlobalFree(filelist.list);
	filelist.list = (LPTSTR)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
		(filelist.len + 1) * filelist.num * sizeof(TCHAR));
	filelist.frm = k = 0;
	lpList = filelist.list;
	hFind = FindFirstFile(szTempBuf, &FindData);
	if (hFind == INVALID_HANDLE_VALUE) return (-1);
	do {
		_tcscpy_s(lpList, filelist.len + 1, FindData.cFileName);
		if (_tcscmp(filelist.szBaseName, lpList) == 0) {
			filelist.frm = k;
		}
		k++;
		lpList += filelist.len + 1;
	} while (FindNextFile(hFind, &FindData));
	FindClose(hFind);

	return (0);
}

#define EOTF(x) ((x < 0.08145)? (x / 4.5) : (pow((x + 0.0993) / 1.0993, 1.0 / 0.45)))
/* #define OETF(x) ((x < 0.0181)? (x * 4.5) : (1.0993 * pow(x, 0.45) - 0.0993)) */
static double lut_eotf[256];  /* LUT for Electro-Optical Transfer Function */
static double lut_oetf[256];  /* LUT for Opto-Electronic Transfer Function */

void InitLUTs()
{
	int idx;
	double e;

	for (idx = 0; idx < 256; idx++) {
		e = (idx - 16) / 219.0;
		e = EOTF(e);
		lut_eotf[idx] = e;
		e = (idx - 16.5) / 219.0;
		e = EOTF(e);
		lut_oetf[idx] = e;
	}
	return;
}

/* Inverse Quantization and Electro-Optical Transfer Function */
double IqEotf(int idx)
{
	return (lut_eotf[idx]);
}

/*  Opto-Electronic Transfer Function and Quantization */
int OetfQt(double e)
{
	int step, idx = 128;

	/* Inverse table look-up by binary search */
	for (step = 64; step > 0; step >>= 1) {
		if (e < lut_oetf[idx]) {
			idx -= step;
		}
		else {
			idx += step;
		}
	}
	if (e < lut_oetf[idx]) idx--;
	return (idx);
}

/* Color conversion */
void BT2020toBT709(int *rgb)
{
	const double mat[3][3] = {
		{ 1.6605, -0.5876, -0.0729 },
		{ -0.1244,  1.1328, -0.0084 },
		{ -0.0181, -0.1006,  1.1188 }
	};
	int i, j;
	double rgb2020[3], rgb709[3];

	for (i = 0; i < 3; i++) {
		rgb2020[i] = IqEotf(rgb[i]);
	}
	for (i = 0; i < 3; i++) {
		rgb709[i] = 0.0;
		for (j = 0; j < 3; j++) {
			rgb709[i] += mat[i][j] * rgb2020[j];
		}
	}
	for (i = 0; i < 3; i++) {
		rgb[i] = OetfQt(rgb709[i]);
	}
	return;
}

DWORD GetDword(HANDLE hFile)
{
	BYTE buf[4];
	DWORD dwRead;

	ReadFile(hFile, buf, 4, &dwRead, NULL);
	return ((buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]);
}

int ReadDPX()
{
	HANDLE hFile;
	DWORD dwRead, dwOffset;
	LPRGBQUAD lpRGB;
	int i, skip, x, y;
	int rgb[3];

	hFile = CreateFile(filelist.szFileName, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return (-1);
	if (GetDword(hFile) != 0x53445058) {	/* magic number "SDPX" */
		CloseHandle(hFile);
		return (-1);
	}
	dwOffset = GetDword(hFile);
	SetFilePointer(hFile, 772, NULL, FILE_BEGIN);
	x = GetDword(hFile);
	y = GetDword(hFile);
	if (x != width && y != height) {
		CloseHandle(hFile);
		if (x > W8K || y > H8K) {
			return (-1);
		}
		width = x;
		height = y;
		return (0);
	}
	SetFilePointer(hFile, dwOffset, NULL, FILE_BEGIN);
	skip = 1 << scale;
	for (y = height - 1; y >= 0; y--) {
		if (y % skip != 0) {
			SetFilePointer(hFile, width * 6, NULL, FILE_CURRENT);
			continue;
		}
		ReadFile(hFile, lpScanline, width * 6, &dwRead, NULL);
		lpRGB = (LPRGBQUAD)lpPixel + (y >> scale) * (width >> scale);
		for (x = 0; x < width; x += skip) {
			for (i = 0; i < 3; i++) {
				rgb[i] = lpScanline[x * 6 + i * 2];
			}
			if (colorconv) {
				BT2020toBT709(rgb);
			}
			lpRGB->rgbRed = rgb[0];
			lpRGB->rgbGreen = rgb[1];
			lpRGB->rgbBlue = rgb[2];
			lpRGB++;
		}
	}
	CloseHandle(hFile);
	return (0);
}

int ReadRawRGB()
{
	HANDLE hFile[3];
	DWORD dwRead;
	LPRGBQUAD lpRGB;
	int len, widthStep, i, skip, x, y;
	int rgb[3];

	len = _tcslen(filelist.szFileName);
	if (filelist.szFileName[len - 3] != TEXT('K')) {
		return(-1);
	}
	if (filelist.szFileName[len - 4] == TEXT('4')) {
		if (width != W4K || height != H4K) {
			width = W4K;
			height = H4K;
			return (0);
		}
	}
	if (filelist.szFileName[len - 4] == TEXT('8')) {
		if (width != W8K || height != H8K) {
			width = W8K;
			height = H8K;
			return (0);
		}
	}
	for (i = 0; i < 3; i++) {
		filelist.szFileName[len - 1] = TEXT("rgb")[i];
		hFile[i] = CreateFile(filelist.szFileName, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (hFile[i] == INVALID_HANDLE_VALUE) {
			while (--i >= 0) {
				CloseHandle(hFile[i]);
			}
			return (-1);
		}
	}
	skip = 1 << scale;
	widthStep = width * 2;
	for (y = height - 1; y >= 0; y--) {
		if (y % skip != 0) {
			for (i = 0; i < 3; i++) {
				SetFilePointer(hFile[i], widthStep, NULL, FILE_CURRENT);
			}
			continue;
		}
		for (i = 0; i < 3; i++) {
			ReadFile(hFile[i], lpScanline + widthStep * i, widthStep, &dwRead, NULL);
		}
		lpRGB = (LPRGBQUAD)lpPixel + (y >> scale) * (width >> scale);
		for (x = 0; x < width; x += skip) {
			for (i = 0; i < 3; i++) {
				rgb[i] = ((lpScanline[widthStep * i + x * 2] << 4) |
						  (lpScanline[widthStep * i + x * 2 + 1] >> 4)) & 0xff;
			}
			if (colorconv) {
				BT2020toBT709(rgb);
			}
			lpRGB->rgbRed = rgb[0];
			lpRGB->rgbGreen = rgb[1];
			lpRGB->rgbBlue = rgb[2];
			lpRGB++;
		}
	}
	for (i = 0; i < 3; i++) {
		CloseHandle(hFile[i]);
	}
	return (0);
}

int ReadImage()
{
	int i;

	i = filelist.frm;
	if (i >= 0 && i < filelist.num) {
		_tcscpy_s(filelist.szBaseName, filelist.len + 1,
			filelist.list + (filelist.len + 1) * i);
	} else {
		return (-1);
	}
	if (filelist.isdpx) {
		return (ReadDPX());
	}
	else {
		return (ReadRawRGB());
	}
}

void InitImage()
{
	if (GetFileList()) {
		MessageBox(NULL, TEXT("Can't open file!"),
			"Error", MB_OK);
		filelist.num = filelist.step = 0;
		return;
	}
	if (ReadImage()) {
		MessageBox(NULL, TEXT("Invalid image file!"),
			"Error", MB_OK);
		filelist.num = 0;
	}
	filelist.step = 0;
}

BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT message,
						   WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
		case IDCANCEL:
			EndDialog(hDlg, FALSE);
		}
		return TRUE;
	}
	return FALSE;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wp, LPARAM lp) {
	HDC hdc;
	PAINTSTRUCT ps;
	static HWND hStatus;
	static HMENU hMenu;
	static HINSTANCE hInstance;
	MENUITEMINFO menuinfo;
	HDROP hDrop;
	int right[] = {100, 200, -1}; 
	RECT rc;
	int k;

	switch (msg) {
	case WM_CREATE:
		hStatus = CreateStatusWindow(
			WS_CHILD | WS_VISIBLE |	CCS_BOTTOM | SBARS_SIZEGRIP,
			TEXT("hoge"), hWnd, 1);
		GetWindowRect(hStatus, &rc);
		k = rc.bottom - rc.top;
		SetRect(&rc, 0, 0, bmpInfo.bmiHeader.biWidth, bmpInfo.bmiHeader.biHeight);
		AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, TRUE);
		SetWindowPos(hWnd, NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top + k,
			SWP_NOMOVE|SWP_NOZORDER);
		hMenu = GetMenu(hWnd);
		hInstance = ((LPCREATESTRUCT)lp)->hInstance;
		return 0;
	case WM_SIZE:
		SendMessage(hStatus, WM_SIZE, wp, lp);
		GetWindowRect(hStatus, &rc);
		right[0] = rc.right - rc.left - 250;
		right[1] = rc.right - rc.left - 110;
		SendMessage(hStatus, SB_SETPARTS, 3, (LPARAM)right);
		return 0;
	case WM_COMMAND:
		switch (LOWORD(wp)) {
		case IDM_FILE_OPEN:
			QueryFileName(hWnd);
			InitImage();
			if (bmpInfo.bmiHeader.biWidth != (width >> scale) ||
				bmpInfo.bmiHeader.biHeight != (height >> scale)) {
				goto resize;
			}
			break;
		case IDM_ORIGINAL_SIZE:
		case IDM_HALF_SIZE:
		case IDM_QUARTER_SIZE:
		case IDM_EIGHTH_SIZE:
			menuinfo.cbSize = sizeof(MENUITEMINFO);
			menuinfo.fMask = MIIM_STATE;
			menuinfo.fState = MFS_UNCHECKED;
			SetMenuItemInfo(hMenu, IDM_ORIGINAL_SIZE + scale, FALSE, &menuinfo);
			menuinfo.fState = MFS_CHECKED;
			SetMenuItemInfo(hMenu, LOWORD(wp), FALSE, &menuinfo);
			scale = LOWORD(wp) - IDM_ORIGINAL_SIZE;
resize:
			InitDIB();
			GetWindowRect(hStatus, &rc);
			k = rc.bottom - rc.top;
			SetRect(&rc, 0, 0, bmpInfo.bmiHeader.biWidth, bmpInfo.bmiHeader.biHeight);
			AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, TRUE);
			SetWindowPos(hWnd, NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top + k,
				SWP_NOMOVE|SWP_NOZORDER);
			filelist.step = 0;
			ReadImage();
			InvalidateRect(hWnd, NULL, FALSE);
			UpdateWindow(hWnd);
			break;
		case IDM_COLOR_CONVERSION:
			menuinfo.cbSize = sizeof(MENUITEMINFO);
			menuinfo.fMask = MIIM_STATE;
			if (colorconv) {
				menuinfo.fState = MFS_UNCHECKED;
				colorconv = 0;
			} else {
				menuinfo.fState = MFS_CHECKED;
				colorconv = 1;
			}
			SetMenuItemInfo(hMenu, LOWORD(wp), FALSE, &menuinfo);
			ReadImage();
			break;
		case IDM_FILE_EXIT:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			return (0);
		case IDM_HELP_ABOUT:
			DialogBox(hInstance, "ABOUTDLG", hWnd, AboutDlgProc);
			break;
		default:
			MessageBeep(0);
			return (0);
		}
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		return (0);
	case WM_KEYDOWN:
		switch(wp) {
		case VK_RIGHT:
			if (filelist.step < 10) filelist.step++;
			break;
		case VK_LEFT:
			if (filelist.step > -10) filelist.step--;
			break;
		case 'Q':
			PostQuitMessage(0);
			return (0);
		default:
			filelist.step = 0;
			break;
		}
		if (filelist.num == 0) filelist.step = 0;
		if (filelist.step == 0) {
			KillTimer(hWnd, 1);
		} else {
			SetTimer(hWnd, 1, 1000 / 60, NULL);
		}
		wsprintf(szTempBuf, "Step: %s%d",
			(filelist.step > 0)? TEXT("+"): TEXT(""), filelist.step);
		SendMessage(hStatus, SB_SETTEXT, 2, (LPARAM)szTempBuf);
		return (0);
	case WM_TIMER:
		filelist.frm += filelist.step;
		if (filelist.frm < 0) filelist.frm = filelist.num - 1;
		if (filelist.frm >= filelist.num) filelist.frm = 0;
		ReadImage();
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		return (0);
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		SetDIBitsToDevice(hdc, 0, 0,
			bmpInfo.bmiHeader.biWidth, bmpInfo.bmiHeader.biHeight,
			0, 0, 0, bmpInfo.bmiHeader.biHeight,
			lpPixel, &bmpInfo, DIB_RGB_COLORS);
		SendMessage(hStatus, SB_SETTEXT, 0, (LPARAM)filelist.szBaseName);
		wsprintf(szTempBuf, "Frame: %d/%d", filelist.frm, filelist.num);
		SendMessage(hStatus, SB_SETTEXT, 1, (LPARAM)szTempBuf);
		wsprintf(szTempBuf, "Step: %s%d",
			(filelist.step > 0)? TEXT("+"): TEXT(""), filelist.step);
		SendMessage(hStatus, SB_SETTEXT, 2, (LPARAM)szTempBuf);
		EndPaint(hWnd, &ps);
		return (0);
	case WM_DROPFILES:
		hDrop = (HANDLE)wp;
		DragQueryFile(hDrop, 0, filelist.szFileName, MAX_PATH);
		DragFinish(hDrop);
		InitImage();
		if (bmpInfo.bmiHeader.biWidth != (width >> scale) ||
			bmpInfo.bmiHeader.biHeight != (height >> scale)) {
			goto resize;
		}
		InvalidateRect(hWnd, NULL, FALSE);
		UpdateWindow(hWnd);
		return 0;
	case WM_DESTROY:
		PostQuitMessage (0);
		return (0);
	}
	return DefWindowProc(hWnd, msg, wp, lp);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   LPSTR lpCmdLine, int nCmdShow)
{
	HWND hWnd;
	MSG msg;
	WNDCLASS winc;

	InitDIB();
	filelist.list = NULL;
	filelist.num = filelist.step = 0;
	winc.style			= CS_HREDRAW | CS_VREDRAW;
	winc.lpfnWndProc	= WndProc;
	winc.cbClsExtra		= winc.cbWndExtra = 0;
	winc.hInstance		= hInstance;
	winc.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	winc.hCursor		= LoadCursor(NULL, IDC_ARROW);
	winc.hbrBackground	= (HBRUSH)GetStockObject(GRAY_BRUSH);
	winc.lpszMenuName	= "MAINMENU";
	winc.lpszClassName	= TEXT("UHDviewer");
	if (*lpCmdLine) {
		if (*lpCmdLine == TEXT('\"')) lpCmdLine++;
		_tcscpy_s(filelist.szFileName, MAX_PATH, lpCmdLine);
		InitImage();
		if (bmpInfo.bmiHeader.biWidth != (width >> scale) ||
			bmpInfo.bmiHeader.biHeight != (height >> scale)) {
			InitDIB();
			ReadImage();
		}
	}
	InitLUTs();

	if (!RegisterClass(&winc)) return 1;

	hWnd = CreateWindow(
		TEXT("UHDviewer"), TEXT("4K/8K Test Sequence Viewer"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT,
        bmpInfo.bmiHeader.biWidth, bmpInfo.bmiHeader.biHeight,
		NULL, NULL, hInstance, NULL
	);
	if (hWnd == NULL) return (1);

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
	DragAcceptFiles(hWnd, TRUE);
	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}
